#include "App.h"

#include <cstdlib>
#include <iostream>
#include <typeinfo>
#include <X11/Xutil.h>

////////////////////////////////////////////////////////////////////////////////
App::App(const std::string & appname)
   {
   XSetErrorHandler(&App::x_error_handler);

   if ((dpy = XOpenDisplay(NULL)) == NULL) 
      {
      std::cerr << "Unable to connect to display " << XDisplayName(NULL) << std::endl;
      exit(1);
      }
   std::cout << "Display name     : " << XDisplayName(NULL) << std::endl
	     << "Number of screens: " << ScreenCount(dpy) << std::endl
	     << "Default screen   : " << DefaultScreen(dpy) << std::endl
	     << "Protocol Version : " << ProtocolVersion(dpy) << std::endl
	     << "Protocol Revision: " << ProtocolRevision(dpy) << std::endl
	     << "Server Vendor    : " << ServerVendor(dpy) << std::endl
	     << "Vendor Release   : " << VendorRelease(dpy) << std::endl
	     << std::endl
	     << "Display width    : " << this->display_width() << std::endl
	     << "Display height   : " << this->display_height() << std::endl
	 ;

   App::BitMap * icon = new App::BitMap("xlib_App.xbm");

   // Create the main window
   create_window("main", 
		 0, 0 , 
		 this->display_width()/3, this->display_height()/4, 
		 icon);

   // Even though we are not passing any actual size_hints, the Xlib programming
   // manual says to pass the flags below anyway.
   XSizeHints *size_hints;
   if (!(size_hints = XAllocSizeHints())) 
      {
      std::cerr << "XAllocSizeHints: memory allocation failure" << std::endl;
      exit(1);
      }
   size_hints->flags = PPosition | PSize;

   XTextProperty windowName, iconName;
   const char * ptr = appname.c_str();
   if (XStringListToTextProperty(const_cast<char **>(&ptr), 1, &windowName) == 0) 
      {
      std::cerr << "XStringListToTextProperty: structure allocation for windowName failed." << std::endl;;
      exit(1);
   }
   ptr = icon->name.c_str();
   if (XStringListToTextProperty(const_cast<char **>(&ptr), 1, &iconName) == 0) 
      {
      std::cerr << "XStringListToTextProperty: structure allocation for iconName failed." << std::endl;;
      exit(1);
   }

   XWMHints *wm_hints;
   if (!(wm_hints = XAllocWMHints()))
      {
      std::cerr << "XAllocWMHints: memory allocation failure" << std::endl;
      exit(0);
      }
   wm_hints->initial_state = NormalState;
   wm_hints->input = True;
   wm_hints->icon_pixmap = icon->pixmap;
   wm_hints->flags = StateHint | IconPixmapHint | InputHint;

   App::WindowData * wd = windows["main"];
    XSetWMProperties(dpy, wd->win, &windowName, &iconName,
   		    NULL, 0, size_hints, wm_hints,
   		    NULL);
   }

////////////////////////////////////////////////////////////////////////////////
App::~App()
   {
   for( WindowMap::iterator i = windows.begin(); i != windows.end(); ++i)
      {
      delete (*i).second->icon;
      App::WindowData * wp = (*i).second;
      windows.erase(i);
      delete wp;
      }

   if (dpy)
      {
      XCloseDisplay(dpy);
      }
   }

////////////////////////////////////////////////////////////////////////////////
int App::x_error_handler(Display * dpy, XErrorEvent *err)
   {
   char err_msg[1024];
   err_msg[0] = 0;
   XGetErrorText(dpy, err->error_code, err_msg, 1024);
   std::cerr << "X11 Error"<< std::endl
	     << "    display     : " << DisplayString(dpy) << std::endl
	     << "    serial      : " << err->serial << std::endl
	     << "    error_code  : " << (int) err->error_code << std::endl
	     << "    request_code: " << (int) err->request_code << std::endl
	     << "    minor_code  : " << (int) err->minor_code << std::endl
	     << err_msg << std::endl;
   return 0;
   }

////////////////////////////////////////////////////////////////////////////////
void App::create_window(const std::string & win_name,
			int x, int y,
			unsigned int width, unsigned int height,
			App::BitMap * icon)
   {
   Window win;
   int border_width = 4;           /* Border four pixels wide */

   int screen_num = DefaultScreen(dpy);

   /* Create opaque window */
   win = XCreateSimpleWindow(dpy, 
			     RootWindow(dpy, screen_num),
			     x, y, 
			     width, height, 
			     border_width, 
			     BlackPixel(dpy, screen_num), 
			     WhitePixel(dpy, screen_num));

   icon->make_pixmap(dpy, win);

   // Add it to the window list
   App::WindowData * win_data = new App::WindowData;
   win_data->name = win_name;
   win_data->win = win;
   win_data->icon = icon;
   windows.insert(win_pair(win_name, win_data));
   }

////////////////////////////////////////////////////////////////////////////////
// App::BitMap
////////////////////////////////////////////////////////////////////////////////
App::BitMap::BitMap(const std::string & file)
   {
   read_from_file(file);
   }

////////////////////////////////////////////////////////////////////////////////
App::BitMap::~BitMap() 
   {
   if (data) XFree(data);
   }

////////////////////////////////////////////////////////////////////////////////
void App::BitMap::make_pixmap(Display * dpy, Drawable d)
   {
   pixmap = XCreateBitmapFromData(dpy, d, (char *) data, width, height);
   }

////////////////////////////////////////////////////////////////////////////////
int App::BitMap::read_from_file(const std::string & file)
   {
   if (BitmapSuccess != XReadBitmapFileData(file.c_str(), &width, &height, &data, NULL, NULL))
      {
      return 0;
      }

   name = file;
   size_t p = name.rfind("/");
   if (p != std::string::npos)
      {
      name.replace(0, p+1, "");
      }
   return 1;
   }

